/*
 *  Routines for determining a lower bound on the cost of an instance
 *  of the combinatorial optimization problem for learning a Bayesian
 *  network from data.
 */
#ifndef _LOWERBOUND_H_
#define _LOWERBOUND_H_

class Instance;
class Graph;
class Pattern;
class LocalSearch;

class Bound
{
    public:
	Bound( Instance *instance );
	~Bound();

	/*
	 *  Greedily construct a lower bound on the cost of completing
	 *  a partial solution by picking the lowest cost domain element
	 *  for each variable. Relies on the domains being sorted by
	 *  increasing costs.
	 */
	Score getGreedyLowerBound( bool *instantiated );

	/*
	 *  Greedily construct an upper bound on the cost of completing
	 *  a partial solution by picking the highest cost domain element
	 *  for each variable. Relies on the domains being sorted by
	 *  increasing costs.
	 */
	Score getGreedyUpperBound( bool *instantiated );

	/*
	 *  Construct a lower bound on the cost of completing
	 *  a partial solution using static partitioning.
	 *
	 *  The idea is to partition the vertices into blocks
	 *  of fixed size and find the optimal solution within
	 *  each partitiion using dynamic programming.
	 */
	Score getLowerBound( bool *instantiated );

	/*
	 *  Construct an upper bound on the best solution using
	 *  local search.
	 */
	Score getUpperBound(
		int *bestOrdering, int *bestSolution, int strength );

	/*
	 *  Construct the partitions and pattern databases for
	 *  the lower bound computation.
	 */
	void partition( bool *instantiated );
	void repartition( bool *instantiated );

    private:
	Instance *instance;
	Graph *graph;

	/*
	 *  Routines for finding upper bounds.
	 */
	LocalSearch *localSearch;

	/*
	 *  Routines and data structures for finding lower bounds.
	 *  The idea is to partition the vertices into blocks and
	 *  find the optimal solution within each partitiion using
	 *  dynamic programming.
	 *
	 *    MAX_TOP: use top = 1, ..., MAX_TOP of lowest cost parent sets
	 *    MAX_ELEM_SMALL: max elements in pattern db for partition
	 *    MAX_ELEM_LARGE: max elements in pattern db for repartition
	 */
	static const int MAX_TOP = 10;
	static const int MAX_ELEM_SMALL = 20;
	static const int MAX_ELEM_LARGE = 26;
	Pattern **pattern;
	void partitionAux( bool *instantiated, int maxPartitionSize );
	Score fillBlock( int i );
	Score fillPartition();
	void improvePartition();
	void coalescePartition( int maxPartitionSize );
	int block[MAX_VARS][MAX_ELEM_LARGE]; // i^th block's elements
	int last[MAX_VARS];		     // i^th block's number of elements
	int nBlocks;			     // number of blocks in partition
};

#endif // _LOWERBOUND_H_
