/*
 *  Routines for enforcing the constraint that the parent
 *  variables must form a directed acyclic graph.
 */

#include "Instance.h"
#include "DAG.h"

/*
 *  Constructor for DAG.
 */
DAG::DAG( Instance *instance )
{
    this->instance = instance;
}

/*
 *  Destructor for DAG.
 */
DAG::~DAG()
{
}

/*
 *  Propagate the acyclic constraint over the parent variables.
 */
bool
DAG::propagate(
	int *currentParents, int *currentOrdering, int *currentDepth,
	int currentLevel, bool *instantiatedP, bool **checkingP, int v )
{
    bool success;
    success = propagateNecessaryEdges( currentLevel, instantiatedP, checkingP, v );
    if( !success ) return( false );

    success = isAcyclic( currentOrdering, currentLevel );
    if( !success ) return( false );

    /*
     *  Prune low cost parent sets based on acyclic constraint.
     *  Does not seem to improve performance under various parameter settings.
     */
/*
    for( int i = 0; i < instance->nVars(); i++ ) {
	if( !instantiatedP[i] && (instance->getCurrentDomainSizeP( i ) < 5) ) {
	    int nTest = 0;
	    for( int a = 0; a <= instance->getLastValidP( i ); a++ ) {
	        if( instance->validP( i, a ) ) {
		    if( !isAcyclic( currentOrdering, currentLevel, i, a ) ) {
		        //printf("(l) prune %d from domain of parent variable %d\n",a,i);
		        instance->setValidP( i, a, currentLevel );
		        checkingP[v][i] = true;
		        if( instance->getCurrentDomainSizeP( i ) == 0 ) {
		            return( false );
		        }
		    }
		    nTest++;
		    if( nTest >= 3 ) break;
	        }
	    }
        }
    }
*/

    return( true );
}


/*
 *  Pruning rule based on necessary edges.
 *
 *  If a directed edge i --> j is a necessary edge,
 *  can remove any parent set that contains j --> i
 *  from the domain of (the uninstantiated variable) i.
 *  An edge i --> j is necessary if i occurs in every
 *  currently valid parent set for variable j.
 *
 *  Determining something more general---reachability in the graph
 *  of necessary edges---is not worthwhile as the difference in
 *  pruning was negligible and as pruning progresses new edges
 *  become necessary (so there is the expense of maintaining
 *  the graph of necessary edges).
 */
bool
DAG::propagateNecessaryEdges(
	int currentLevel, bool *instantiatedP, bool **checkingP, int v )
{
    instance->findNecessaryParents();
    for( int i = 0; i < instance->nVars(); i++ ) if( !instantiatedP[i] ) {
	/*
	 *  Set necessary is the set of all j such that i --> j is a
	 *  necessary edge.
	 */
	Set necessary = instance->necessaryEdges( i );
	if( empty( necessary ) ) continue;
	for( int a = 0; a <= instance->getLastValidP( i ); a++ ) {
	    if( instance->validP( i, a ) ) {
		/*
		 *  Remove the parent set a from domain of i if the
		 *  intersection of necessary and a is not empty.
		 */
		Set parent = instance->getParentSet( i, a );
		if( !empty( intersection( necessary, parent ) ) ) {
		    //printf("(j) prune %d from domain of parent variable %d\n",a,i);
		    instance->setValidP( i, a, currentLevel );
		    checkingP[v][i] = true;
		    instance->findNecessaryParents( i );
		    if( instance->getCurrentDomainSizeP( i ) == 0 ) {
		        return( false );
		    }
		}
	    }
	}
    }

    return( true );
}

/*
 *  Satisfiability test.
 *
 *  For every subset C of the variables, there must exist a
 *  v in C such that some of v's parents are not in C.
 */
bool
DAG::isAcyclic( int *currentOrdering, int currentLevel )
{
    Set inOrdering = emptySet;
    for( int i = 0; i <= currentLevel; i++ ) {
	add_element( currentOrdering[i], inOrdering );
    }

    int k = currentLevel+1;
    bool changes;
    do {
	changes = false;
	for( int w = 0; w < instance->nVars(); w++ ) if( !element_of( w, inOrdering ) ) {
	    for( int a = 0; a <= instance->getLastValidP( w ); a++ ) {
		if( instance->validP( w, a ) ) {
		    Set parent = instance->getParentSet( w, a );
		    if( subset_of( parent, inOrdering ) ) {
			add_element( w, inOrdering );
			k++;
			changes = true;
			break;
		    }
		}
	    }
	}
    } while( changes );

    if( k != instance->nVars() ) {
	//printf("(k) cycle detected, current level = %d\n", currentLevel);
	return( false );
    }

    return( true );
}

/*
 *  Satisfiability test. Alternate version.
 *  This version is quite a bit slower than isAcyclic()
 *  but could be used to prune the ordering and
 *  depth variables.
 *
 *  For every subset C of the variables, there must exist a
 *  v in C such that some of v's parents are not in C.
 */
bool
DAG::isAcyclic( int *currentOrdering, int *currentDepth, int currentLevel )
{
    int ordering[instance->nVars()];
    int depth[instance->nVars()];
    int cd = currentDepth[currentLevel];
    int k = currentLevel+1;
    for( int i = 0; i < k; i++ ) {
	ordering[i] = currentOrdering[i];
	depth[i] = currentDepth[i];
    }
    bool inOrdering[instance->nVars()];
    for( int i = 0; i < instance->nVars(); i++ ) {
	inOrdering[i] = false;
    }
    for( int i = 0; i < k; i++ ) {
	int w = ordering[i];
	inOrdering[w] = true;
    }

    /*
     *  Step 1. Complete finding all the remaining w not
     *          yet in the ordering that belong at depth d.
     *          A variable w belongs at depth d if there exists a
     *          p in dom(w) such that depth( p | ... ) = d
     */
    int m = k;
    for( int w = 0; w < instance->nVars(); w++ ) if( !inOrdering[w] ) {
	for( int a = 0; a <= instance->getLastValidP( w ); a++ ) {
	    if( instance->validP( w, a ) ) {
		int d = instance->getDepthParentSet( ordering, depth, m, w, a );
		if( d == cd ) {
		    ordering[k] = w;
		    depth[k] = cd;
		    inOrdering[w] = true;
		    k++;
		    break;
		}
	    }
	}
    }
    m = k;
    /*
     *  Step 2. d++
     *          while not all of the variables are in the ordering
     *		    find all w not yet in ordering that belong at depth d
     *		    if there are no such variables, return false
     *		    d++
     */
    cd++;
    while( m < instance->nVars() ) {
	for( int w = 0; w < instance->nVars(); w++ ) if( !inOrdering[w] ) {
	    for( int a = 0; a <= instance->getLastValidP( w ); a++ ) {
		if( instance->validP( w, a ) ) {
		    int d = instance->getDepthParentSet(
				ordering, depth, m, w, a );
		    if( d == cd ) {
		        ordering[k] = w;
		        depth[k] = cd;
		        inOrdering[w] = true;
		        k++;
		        break;
		    }
		}
	    }
	}

	/*
	 *  If the depth layer is empty, then unsatisfiable.
	 */
	if( m == k ) {
	    //printf("(k) cycle detected, current level = %d\n", currentLevel);
	    return( false );
	}

	m = k;
	cd++;
    }

    return( true );
}

/*
 *  Satisfiability test when v = a.
 *
 *  For every subset C of the variables, there must exist a
 *  v in C such that some of v's parents are not in C.
 */
bool
DAG::isAcyclic( int *currentOrdering, int currentLevel, int v, int a )
{
    Set inOrdering = emptySet;
    for( int i = 0; i <= currentLevel; i++ ) {
	add_element( currentOrdering[i], inOrdering );
    }

    int k = currentLevel+1;
    bool changes;
    do {
	changes = false;
	if( !element_of( v, inOrdering ) ) {
	    Set parent = instance->getParentSet( v, a );
	    if( subset_of( parent, inOrdering ) ) {
		add_element( v, inOrdering );
		k++;
		changes = true;
	    }
	}
	for( int w = 0; w < instance->nVars(); w++ ) if( (w != v) && !element_of( w, inOrdering ) ) {
	    for( int a = 0; a <= instance->getLastValidP( w ); a++ ) {
		if( instance->validP( w, a ) ) {
		    Set parent = instance->getParentSet( w, a );
		    if( subset_of( parent, inOrdering ) ) {
			add_element( w, inOrdering );
			k++;
			changes = true;
			break;
		    }
		}
	    }
	}
    } while( changes );

    if( k != instance->nVars() ) {
	return( false );
    }

    return( true );
}

