/*
 *  Instance: An instance of the combinatorial optimization problem
 *	      for learning a Bayesian network from data.
 */
#ifndef _INSTANCE_H_
#define _INSTANCE_H_

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

/*
 *  Definitions for scores.
 *
 *  Defines are in case the type definition changes.
 *  Using floats or doubles instead is possible but
 *  testing equality or even less than is problemetic.
 *  Would need something along these lines:
 *      static const double epsilon = 1.0e-05;
 *      #define lessThan(a,b) ((a - b) < -epsilon)
 */
typedef uint64_t Score;
#define gt(a,b) (a > b)
#define ge(a,b) (a >= b)
#define lt(a,b) (a < b)
#define le(a,b) (a <= b)
static const Score MAX_SCORE = 99999999999999;// larger than any possible score
static const double scale = 100000.0;	// 5 digits of precision
static const int precision = 5;		// 5 digits of precision

/*
 *  Definitions for parent sets.
 */
typedef uint64_t Set;
#define empty(s) (s == 0)
#define complement(s) (~s)
#define add_element(e,s) (s |= ((Set)1 << e))
#define remove_element(e,s) (s ^= ((Set)1 << e))
#define element_of(e,s) (s & ((Set)1 << e))
#define subset_of(s,t) ((s & t) == s)
#define intersection(s,t) (s & t)
#define union(s,t) (s | t)
#define equal(s,t) (s == t)
#define emptySet ((Set)0x0000000000000000)
#define universalSet ((Set)0xFFFFFFFFFFFFFFFF)
static const int MAX_VARS = 64;	// sizeof(Set)

class Graph;

class Instance
{
    public:
	Instance( char *fileName );
	~Instance();

	/*
	 *  Number of random variables in instance.
	 */
	int nVars();

	/*
	 *  Domain size of
	 *	parent variable v,
	 *	ordering variable v,
	 *	depth variable v.
	 */
	int getDomainSizeP( int v );
	int getDomainSizeO( int v );
	int getDomainSizeD( int v );

	/*
	 *  Current domain size of
	 *	parent variable v,
	 *	ordering variable v,
	 *	depth variable v.
	 */
	int getCurrentDomainSizeP( int v );
	int getCurrentDomainSizeO( int v );
	int getCurrentDomainSizeD( int v );
	int getLastValidP( int v );

	/*
	 *  Name of instance.
	 */
	char *benchmarkName();

	/*
	 *  Determine the minimum cost subnetwork (solution) and its cost
	 *  for the ordering, given the ancestors.
	 */
	Score minCostOrdering( Set ancestors,
		int *ordering, int m, int *solution = NULL );

	/*
	 *  Determine the parent set for the variable v that has the
         *  minimum cost given v that comes next in the ordering;
	 *  i.e., all parents of v must be in the ordering or in the
	 *  set of ancestors.
	 */
	int getMinCostParentSet( int *ordering, int m, int v );
	int getMinCostParentSet( Set ancestors, int v );

	/*
	 *  Determine the depth of the parent set a for the variable v given
	 *  that v comes next in the ordering; i.e., all parents a of v will
	 *  exist in the ordering. Ordering is valid for ordering[0..m-1].
	 */
	int getDepthParentSet(
		int *ordering, int *depth, int m, int v, int a );
	int getDepthParentSetUnordered(
		int *ordering, int *depth, int m, int v, int a );

	/*
	 *  Given parents set a of variable v, determine whether there
	 *  exists a parent set b that is of lower cost and is contained
	 *  in the induced parent set of a; i.e., where it is known that
	 *  all of the variables in the ordering precede variable v.
	 */
	bool containedInIP( int *ordering, int m, int v, int a );

	/*
	 *  Return the cost of domain element a of parent variable v.
	 */
	Score getCost( int v, int a );

	/*
	 *  Determine whether x = a and y = b form a covered edge;
	 *  i.e., x = P and y = (P union { x }).
	 *  Determine whether y = b and z = c form a covered edge;
	 *  if the covered edge x --> y is flipped to be y --> x.
	 */
	bool coveredEdge( int x, int a, int y, int b );
	bool flippedCoveredEdge( int x, int a, int y, int b, int z, int c );
	bool triangleCoveredEdge(
		int w, int a,
		int x, int b,
		int y, int c,
		int z, int d );

	/*
	 *  Print the parent set of element a of variable v.
	 */
	void printParentSet( int v, int a );

	/*
	 *  Print a solution.
	 */
	void printSolution( int *ordering, int *solution, Score cost );

	/*
	 *  Routines for querying and setting whether domain element a of
	 *  variable v is valid. A value of -1 indicates valid;
	 *  0, ..., n-1 indicate the element was removed at that level
	 *  in the search.
	 */
	bool validP( int v, int a );
	bool validO( int v, int a );
	bool validD( int v, int a );
	int  getValidP( int v, int a );
	int  getValidO( int v, int a );
	int  getValidD( int v, int a );
	void setValidP( int v, int a, int value );
	void setValidO( int v, int a, int value );
	void setValidD( int v, int a, int value );

	/*
	 *  Determine set of necessary parents of vertices.
	 *  Determine set of necessary edges for vertices.
	 */
	void findNecessaryParents();
	void findNecessaryParents( int w );
	Set necessaryEdges( int v );
	Set getParentSet( int v, int a );

	/*
	 *  Update parents sets given that parent variables
	 *  v and w are symmetric.
	 */
	void symmetricParentVariables( int v, int w );

	/*
	 *  Add all or just the top few parent sets as
	 *  edges in the graph.
	 */
	void addEdges( Graph *graph );
	void addEdges( Graph *graph, int top );
	void addEdges( Graph *graph, bool *instantiated );

	/*
	 *  Debugging.
	 */
	void dumpInstance();
	void dumpState(
	    int currentLevel,
	    int *currentParents,
	    int *currentOrdering,
	    int *currentDepth,
	    bool *instantiatedP );
	void dumpStateFull(
	    int currentLevel,
	    int *currentParents,
	    int *currentOrdering,
	    int *currentDepth,
	    bool *instantiatedP );

    private:
	void initialize( char *fileName );
	char benchmark[128];

	/*
	 *  Structure of a domain value: The parent set, the cost
	 *  of this parent set, and whether the domain value is valid;
	 *  i.e., has not been pruned away by constraint propagation.
	 */
	typedef struct {
	    Score  cost;
	    Set    parentSet;
	    int    cardinality;
	    int    valid;
	} Value;

	static int compare( const void *a, const void *b );

	/*
	 *  Constraint model.
	 */
	int n;				// number of random variables
	int domainSizeP[MAX_VARS];	// domain size of parent variables
	Value *domainP[MAX_VARS];	// domain of parent variables
	int domainSizeO[MAX_VARS];	// domain size of ordering variables
	int *domainO[MAX_VARS];		// domain of ordering variables
	int domainSizeD[MAX_VARS];	// domain size of depth variables
	int *domainD[MAX_VARS];		// domain of depth variables

	/*
	 *  Current domain sizes and last valid domain element,
	 *  maintained during search.
	 */
	int domainSizeCurrentP[MAX_VARS];
	int domainSizeCurrentO[MAX_VARS];
	int domainSizeCurrentD[MAX_VARS];
	int lastValidP[MAX_VARS];

	/*
	 *  Determine set of necessary parents of vertices.
	 *
	 *  A vertex v is a necessary parent of vertex w if v occurs
	 *  in every currently valid parent set for variable w.
	 */
	Set necessaryParents[MAX_VARS];
};

#endif // _INSTANCE_H_
